/* userhosts - redirect access to /etc/hosts to another file */
/*
 * Copyright (C) 2014 Vincent Pelletier <vincent@nexedi.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.

 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

/* Needed for RTLD_NEXT */
#define _GNU_SOURCE

#include <string.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <dlfcn.h>
#include <fcntl.h>

#define ORIGINAL_HOSTS_PATH "/etc/hosts"

/* Standard O_TMPFILE includes O_DIRECTORY bit, but we do not want to call
 * va_arg if O_DIRECTORY is set alone. Hence, define a "purer" O_TMPFILE. */
#define PURE_O_TMPFILE (020000000)
/* And sanity-check when O_TMPFILE is defined. */
#if defined(O_TMPFILE) && (O_TMPFILE & PURE_O_TMPFILE) == 0
#error PURE_O_TMPFILE is incorrectly defined
#endif

static int (*original_open)(const char *, int, ...);
static FILE *(*original_fopen)(const char *, const char *);
static const char *replacement_hosts;

/* Call dlsym(RTLD_NEXT, name), abort()'ing with informative message to
 * stderr if it cannot be found. */
static inline void *dlsym_or_abort(const char *name) {
  char *error;
  void *symbol;
  dlerror(); /* Clear any previous error */
  symbol = dlsym(RTLD_NEXT, name);
  if (!symbol && (error = dlerror())) {
    fprintf(stderr, "Error loading '%s': %s\n", name, error);
    abort();
  }
  return symbol;
}

static void __attribute__ ((constructor)) init(void) {
  original_open = dlsym_or_abort("open");
  original_fopen = dlsym_or_abort("fopen");
  replacement_hosts = getenv("HOSTS");
  if (!replacement_hosts)
    /* XXX: warn ? fallback on $HOME/???/hosts ? */
    replacement_hosts = ORIGINAL_HOSTS_PATH;
}

int open(const char *__file, int __oflag, ...) {
  if (!strcmp(__file, ORIGINAL_HOSTS_PATH))
    __file = replacement_hosts;
  if (__oflag & (O_CREAT | PURE_O_TMPFILE)) {
    va_list ap;
    mode_t mode;

    va_start(ap, __oflag);
    mode = va_arg(ap, mode_t);
    va_end(ap);
    return (*original_open)(__file, __oflag, mode);
  }
  return (*original_open)(__file, __oflag);
}

FILE *fopen(const char *path, const char *mode) {
  if (!strcmp(path, ORIGINAL_HOSTS_PATH))
    path = replacement_hosts;
  return (*original_fopen)(path, mode);
}
